/*
 * $Id: CGenerator.java,v 1.9 2004/06/23 19:49:31 nunnari Exp $
 *
 * Copyright (c) 2003-2004 Roberto Nunnari - roberto.nunnari@nunnisoft.ch
 * All rights reserved.
 *
 *  This file is part of NunniFSMGen.
 *
 *  NunniFSMGen is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  NunniFSMGen is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with NunniFSMGen; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.nunnisoft.fsmgen;

import java.io.*;
import java.util.*;
import com.nunnisoft.fsmgen.lang.*;

/**
 *
 * @author  Roberto Nunnari
 */
public class CGenerator implements Generator {

    String          m_basename = null;
    String          m_eparamType = null;
    String          m_copy = null;
    String          m_initialState = null;
    String          m_errorState = null;
    Vector          m_entries = null;
    Vector          m_states = null;
    Vector          m_events = null;
    Vector          m_contextMethods = null;
    boolean         m_verbose = false;


    /** Creates a new instance of JavaGenerator */
    public CGenerator( String baseName, String eparamType, String copy, String initialState, String errorState, Vector entries, boolean verbose ) {
        m_basename = baseName;
        m_eparamType = eparamType;
        if ( copy == null )
            m_copy = copy;
        else
            m_copy = copy.trim();
        m_initialState = initialState;
        m_errorState = errorState;
        m_verbose = verbose;
        initFromTransitions( entries );
    }

    private void initFromTransitions( Vector entries ) {
        m_entries = entries;
        m_states = new Vector( 10, 10 );
        m_events = new Vector( 10, 10 );
        m_contextMethods = new Vector( 10, 10 );
        Iterator iter = entries.iterator();
        while ( iter.hasNext() ) {
            SMTEntry entry = (SMTEntry)iter.next();

            if ( !m_states.contains( entry.m_actualState ) )
                m_states.add( entry.m_actualState );
            if ( !m_events.contains( entry.m_event ) )
                m_events.add( entry.m_event );
            if ( !m_contextMethods.contains( entry.m_action ) )
                m_contextMethods.add( entry.m_action );
        }
    }

    public void generate() throws IOException {
        if ( m_verbose )
            System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GEN_C_SRC ) );
        generateContext();
        generateFSM();
    }



    private void generateContext() throws IOException {
        try {
            String classname = m_basename;
            String filename = classname + ".h";

            File file = new File( filename );
            if ( file.exists() ) {
                System.out.print( "File '" + filename + "' " + ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.ALREADY_EXISTS ) + ": " );
                classname += "New";
                filename = classname + ".h";
                System.out.println( filename + " and " + classname + ".c" );
            }
            PrintWriter out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, true );
            out.println();
            out.println();
            out.println( "#ifndef " + classname + "_H" );
            out.println( "#define " + classname + "_H" );
            out.println();
            out.println();
            out.println( "#ifdef __cplusplus" );
            out.println( "extern \"C\"" );
            out.println( "{" );
            out.println( "#endif		/* __cplusplus */" );
            out.println();
            out.println();
            Iterator iter = m_contextMethods.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                if ( !"-".equals( methodname ) && !"!".equals( methodname ) )
                    out.println( "int " + methodname + "( " + m_eparamType + " o );" );
            }

            out.println();
            out.println();
            out.println( "#ifdef __cplusplus" );
            out.println( "}" );
            out.println( "#endif		/* __cplusplus */" );
            out.println();
            out.println();
            out.println( "#endif" );
            out.close();

            filename = classname + ".c";

            out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, true );
            out.println();
            out.println();
            out.println( "#include \"" + classname + ".h\"" );

            iter = m_contextMethods.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                if ( !"-".equals( methodname ) && !"!".equals( methodname ) ) {
                    out.println();
                    out.println();
                    out.println( "int " + methodname + "( " + m_eparamType + " o ) {" );
                    out.println( "    /* add your code hereafter... */" );
                    out.println( "    return 0;" );
                    out.println( "}" );
                }
            }
            out.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }


    private void generateFSM() throws IOException {
        try {
            String classname = m_basename + "FSM";
            String stateClassname = m_basename + "State";
            String filename = classname + ".h";

            PrintWriter out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, false );
            out.println();
            out.println();
            out.println( "#ifndef " + classname + "_H" );
            out.println( "#define " + classname + "_H" );
            out.println();
            out.println();
            out.println( "#ifdef __cplusplus" );
            out.println( "extern \"C\"" );
            out.println( "{" );
            out.println( "#endif		/* __cplusplus */" );
            out.println();
            out.println();
            out.println( "struct " + classname + ";" );
            out.println();
            out.println();
            out.println( "struct " + classname + " * " + classname + "Create();" );
            out.println();
            out.println();
            out.println( "void " + classname + "Delete( " + "struct " + classname + " *fsm );" );
            out.println();
            out.println( "/* events */" );

            Iterator iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println( "int " + classname + methodname + "( struct " + classname + " *fsm, " + m_eparamType + " o );" );
            }
            out.println();
            out.println();
            out.println( "#ifdef __cplusplus" );
            out.println( "}" );
            out.println( "#endif		/* __cplusplus */" );
            out.println();
            out.println();
            out.println( "#endif" );
            out.close();

            filename = classname + ".c";

            out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, false );
            out.println();
            out.println();
            out.println( "#include \"" + classname + ".h\"" );
            out.println( "#include \"" + m_basename + ".h\"" );
            out.println();
            out.println( "#include <stdlib.h>" );
            out.println();
            out.println();
            out.println( "#define m_maxFSMs 5" );
            out.println();
            out.println();
            out.println( "struct " + stateClassname + " {" );
            iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println( "  int (*" + methodname + ")( struct " + classname + " *fsm, " + m_eparamType + " o );" );
            }
            out.println( "};" );
            out.println();
            out.println();
            out.println( "struct " + classname + " {" );

            iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println( "  int (*" + methodname + ")( struct " + classname + " *fsm, " + m_eparamType + " o );" );
            }

            out.println();
            out.println( "  void (*changeState)( struct " + classname + " *fsm, " + "struct " + stateClassname + " *nextState );" );

            out.println( "  struct " + stateClassname + " *m_state;" );
            out.println( "};" );
            out.println();
            out.println();
            out.println( "static int m_initialized = 0;" );

            iter = m_states.iterator();
            while ( iter.hasNext() ) {
                String cstate = (String)iter.next();
                String cclassname = m_basename + cstate + "State";
                out.println( "static struct " + stateClassname + " m_" + cclassname + ";" );
            }
            out.println();
            out.println();
            out.println( "static struct " + classname + "  * m_FSM[m_maxFSMs];" );
            out.println( "static int m_FSMSize = 0;" );

            iter = m_states.iterator();
            while ( iter.hasNext() ) {
                String cstate = (String)iter.next();
                String cclassname = m_basename + cstate + "State";

                Iterator iter2 = m_entries.iterator();
                while ( iter2.hasNext() ) {
                    SMTEntry entry = (SMTEntry)iter2.next();
                    if ( cstate.equals( entry.m_actualState ) ) {
                        String nextCstate = m_basename + entry.m_nextState + "State";
                        String errorCstate = m_basename + m_errorState + "State";
                        String methodname = entry.m_action;
                        out.println();
                        out.println();
                        out.println( "int " + cclassname + entry.m_event + "( struct " + classname + " *fsm,  " + m_eparamType + " o ) {" );
                        if ( "-".equals( methodname ) ) {
                            if ( !"-".equals( entry.m_nextState ) )
                                out.println( "  fsm->changeState( fsm, &m_" + nextCstate + " );" );
                            out.println( "  return 0;" );
                            out.println( "}" );
                        }
                        else if ( "!".equals( methodname ) ) {
                            out.println( "  fsm->changeState( fsm, &m_" + errorCstate + " );" );
                            out.println( "  return -1;" );
                            out.println( "}" );
                        }
                        else {
                            out.println( "  int ret = " + methodname + "( o );" );
                            out.println( "  if ( ret < 0 )" );
                            out.println( "    fsm->changeState( fsm, &m_" + errorCstate + " );" );
                            if ( !"-".equals( entry.m_nextState ) ) {
                                out.println( "  else" );
                                out.println( "    fsm->changeState( fsm, &m_" + nextCstate + " );" );
                            }
                            out.println( "  return ret;" );
                            out.println( "}" );
                        }
                    }
                }
            }


            iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println();
                out.println();
                out.println( "int " + classname + methodname + "( struct " + classname + " *fsm, " + m_eparamType + " o ) {" );
                out.println( "  return fsm->m_state->" + methodname + "( fsm, o );" );
                out.println( "}" );
            }

            out.println();
            out.println();
            out.println( "void " + classname + "ChangeState( struct " + classname + " *fsm, struct " + stateClassname + " *newState ) {" );
            out.println( "    fsm->m_state = newState;" );
            out.println( "}" );

            // funzione per la creazione di nuove FSMs
            out.println();
            out.println();
            out.println( "struct " + classname + " * " + classname + "Create() {" );
//            out.println( "  int i;" );
            out.println( "  if ( m_FSMSize >= m_maxFSMs )" );
            out.println( "    return 0;" );
            out.println( "  if ( !m_initialized ) {" );
            out.println( "    m_initialized = 1;" );

            iter = m_states.iterator();
            while ( iter.hasNext() ) {
                String cstate = (String)iter.next();
                String cclassname = m_basename + cstate + "State";

                Iterator iter2 = m_entries.iterator();
                while ( iter2.hasNext() ) {
                    SMTEntry entry = (SMTEntry)iter2.next();
                    if ( cstate.equals( entry.m_actualState ) ) {
//                        String nextCstate = m_basename + entry.m_nextState + "State";
                        out.println( "    m_" + cclassname + "." + entry.m_event + " = " + cclassname + entry.m_event + ";" );
                    }
                }
            }
            out.println( "  }" );

            out.println( "  m_FSM[m_FSMSize] = malloc( sizeof(struct " + classname + ") );" );

            iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String event = (String)iter.next();
                Iterator iter2 = m_entries.iterator();
                while ( iter2.hasNext() ) {
                    SMTEntry entry = (SMTEntry)iter2.next();
                    if ( event.equals( entry.m_event ) ) {
                        out.println( "  m_FSM[m_FSMSize]->" + event + " = " + classname + event + ";" );
                        break;
                    }
                }
            }

            out.println( "  m_FSM[m_FSMSize]->changeState = " + classname + "ChangeState;" );
            out.println( "  m_FSM[m_FSMSize]->m_state = &m_" + m_basename + m_initialState + "State;" );
            out.println( "  return (m_FSM[m_FSMSize++]);" );
            out.println( "}" );

            // funzione per cancellare una fsm
            out.println();
            out.println();
            out.println( "void " + classname + "Delete( struct " + classname + " *fsm ) {" );
            out.println( "  int i;" );
            out.println( "  int index = -1;" );
            out.println();
            out.println( "  if ( m_FSMSize == 0 || fsm == NULL )" );
            out.println( "    return;" );
            out.println();
            out.println( "  for ( i = 0; i < m_FSMSize; ++i ) {" );
            out.println( "    if ( m_FSM[i] == fsm ) {" );
            out.println( "      index = i;" );
            out.println( "      break;" );
            out.println( "    }" );
            out.println( "  }" );
            out.println();
            out.println( "  if ( index == -1 )" );
            out.println( "    return;" );
            out.println();
            out.println( "  --m_FSMSize;" );
            out.println();
            out.println( "  for ( i = index ; i < m_FSMSize; ++i ) {" );
            out.println( "    m_FSM[i] = m_FSM[i+1];" );
            out.println( "  }" );
            out.println();
            out.println( "  if ( m_FSMSize == index )" );
            out.println( "    m_FSM[m_FSMSize] = (struct DogFSM *)NULL;" );
            out.println();
            out.println( "  free( fsm );" );
            out.println( "}" );
            out.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }


    private void writeHeader( PrintWriter out, boolean editable ) {
        if ( m_copy != null )
            out.println( m_copy );
        out.println( "/*" );
        if ( !editable )
            out.println( " * this file was generated by NunniFSMGen - do not edit!" );
        else
            out.println( " * this file was generated by NunniFSMGen" );
        out.println( " */" );
    }
}
