/*
 * $Id: CppGenerator.java,v 1.6 2004/06/23 19:49:31 nunnari Exp $
 *
 * Copyright (c) 2003-2004 Roberto Nunnari - roberto.nunnari@nunnisoft.ch
 * All rights reserved.
 *
 *  This file is part of NunniFSMGen.
 *
 *  NunniFSMGen is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  NunniFSMGen is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with NunniFSMGen; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.nunnisoft.fsmgen;

import java.io.*;
import java.util.*;
import com.nunnisoft.fsmgen.lang.*;

/**
 *
 * @author  Roberto Nunnari
 */
public class CppGenerator implements Generator {

    String          m_basename = null;
    String          m_eparamType = null;
    String          m_copy = null;
    String          m_initialState = null;
    String          m_errorState = null;
    Vector          m_entries = null;
    Vector          m_states = null;
    Vector          m_events = null;
    Vector          m_contextMethods = null;
    boolean         m_verbose = false;


    /** Creates a new instance of JavaGenerator */
    public CppGenerator(String baseName, String eparamType, String copy, String initialState, String errorState, Vector entries, boolean verbose) {
        m_basename = baseName;
        m_eparamType = eparamType;
        if ( copy == null )
            m_copy = copy;
        else
            m_copy = copy.trim();
        m_initialState = initialState;
        m_errorState = errorState;
        m_verbose = verbose;
        initFromTransitions( entries );
    }

    private void initFromTransitions( Vector entries ) {
        m_entries = entries;
        m_states = new Vector( 10, 10 );
        m_events = new Vector( 10, 10 );
        m_contextMethods = new Vector( 10, 10 );
        Iterator iter = entries.iterator();
        while ( iter.hasNext() ) {
            SMTEntry entry = (SMTEntry)iter.next();

            if ( !m_states.contains( entry.m_actualState ) )
                m_states.add( entry.m_actualState );
            if ( !m_events.contains( entry.m_event ) )
                m_events.add( entry.m_event );
            if ( !m_contextMethods.contains( entry.m_action ) )
                m_contextMethods.add( entry.m_action );
        }
    }

    public void generate() throws IOException {
        if ( m_verbose )
            System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GEN_CXX_SRC ) );
        generateExceptionClass();
        generateContext();
        generateFSM();
        generateState();
    }



    private void generateExceptionClass() throws IOException {
        try {
            String filename = "LogicError.h";
            PrintWriter out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, false );
            out.println();
            out.println();
            out.println( "#ifndef LogicError_H" );
            out.println( "#define LogicError_H" );
            out.println();
            out.println();
            out.println( "class LogicError {" );
            out.println();
            out.println( "    private :" );
            out.println( "        const char * m_message;" );
            out.println();
            out.println( "    public:" );
            out.println( "        LogicError() : m_message( \"Error!\" ) {}" );
            out.println( "        LogicError( const char *message ) : m_message( message ) {}" );
            out.println( "        const char * getMessage() { return m_message; }" );
            out.println( "};" );
            out.println();
            out.println();
            out.println( "#endif" );
            out.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }


    private void generateContext() throws IOException {
        try {
            String classname = m_basename;
            String filename = classname + ".h";

            File file = new File( filename );
            if ( file.exists() ) {
                System.out.print( "File '" + filename + "' " + ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.ALREADY_EXISTS ) + ": " );
                classname += "New";
                filename = classname + ".h";
                System.out.println( filename + " and " + classname + ".cxx" );
            }
            PrintWriter out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, true );
            out.println();
            out.println();
            out.println( "#ifndef " + classname + "_H" );
            out.println( "#define " + classname + "_H" );
            out.println();
            out.println();
            out.println( "#include \"LogicError.h\"" );
            out.println();
            out.println();
            out.println( "class " + classname );
            out.println( "{" );

            out.println( "    public:" );
            Iterator iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println( "        virtual void " + methodname + "( " + m_eparamType + " o ) throw (LogicError) = 0;" );
            }

            out.println();
            out.println();
            out.println( "    protected:" );
            out.println( "        " + classname + "();" );
            out.println( "        " + classname + "( void *o );" );
            out.println();
            out.println();
            out.println( "    public:" );
            out.println( "        virtual ~" + classname + "();" );

            out.println();
            out.println();
            out.println( "    public:" );
            iter = m_contextMethods.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                if ( !"-".equals( methodname ) && !"!".equals( methodname ) )
                    out.println( "        bool " + methodname + "( " + m_eparamType + " o ) throw (LogicError);" );
            }

            out.println( "};" );
            out.println();
            out.println();
            out.println( "#endif" );
            out.close();

            filename = classname + ".cxx";

            out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, true );
            out.println();
            out.println();
            out.println( "#include \"" + classname + ".h\"" );
            out.println();
            out.println();
            out.println( classname + "::" + classname + "() {}" );
            out.println();
            out.println();
            out.println( classname + "::" + classname + "( void *o ) {}" );
            out.println();
            out.println();
            out.println( classname + "::~" + classname + "() {}" );

            iter = m_contextMethods.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                if ( !"-".equals( methodname ) && !"!".equals( methodname ) ) {
                    out.println();
                    out.println();
                    out.println( "bool " + classname + "::" + methodname + "( " + m_eparamType + " o ) throw (LogicError) {" );
                    out.println( "    // add your code hereafter..." );
                    out.println( "    return true;" );
                    out.println( "}" );
                }
            }
            out.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }


    private void generateFSM() throws IOException {
        try {
            String classname = m_basename + "FSM";
            String stateClassname = m_basename + "State";
            String filename = classname + ".h";

            PrintWriter out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, false );
            out.println();
            out.println();
            out.println( "#ifndef " + classname + "_H" );
            out.println( "#define " + classname + "_H" );
            out.println();
            out.println();
            out.println( "#include \"" + m_basename + ".h\"" );
            out.println();
            out.println();
            out.println( "class " + stateClassname + ";" );
            out.println();
            out.println();
            out.println( "class " + classname + " : public " + m_basename );
            out.println( "{" );
            out.println( "    private:" );
            out.println( "        " + stateClassname + " *m_state;" );
            out.println();
            out.println();
            out.println( "    private:" );
            out.println( "        " + classname + "();" );
            out.println();
            out.println();
            out.println( "    public:" );
            out.println( "        " + classname + "( void *o );" );
            out.println( "        virtual ~" + classname + "();" );
            out.println();
            out.println();
            out.println( "    public:" );
            out.println( "        void changeState( " + stateClassname + " *newState );" );
            Iterator iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println( "        virtual void " + methodname + "( " + m_eparamType + " o ) throw (LogicError);" );
            }
            out.println( "};" );
            out.println();
            out.println();
            out.println( "#endif" );
            out.close();

            filename = classname + ".cxx";

            out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, false );
            out.println();
            out.println();
            out.println( "#include \"" + classname + ".h\"" );
            out.println( "#include \"" + m_basename + "State.h\"" );
            out.println();
            out.println();
            out.println( classname + "::" + classname + "() {}" );
            out.println();
            out.println();
            out.println( classname + "::" +  classname + "( void *o ) : " + m_basename + "( o ) {" );
            out.println( "    m_state = "  + m_basename + m_initialState + "State::instance();" );
            out.println( "}" );
            out.println();
            out.println();
            out.println( classname + "::~" + classname + "() {}" );

            iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println();
                out.println();
                out.println( "void " + classname + "::" + methodname + "( " + m_eparamType + " o ) throw (LogicError) {" );
                out.println( "    m_state->" + methodname + "( this, o );" );
                out.println( "}" );
            }

            out.println();
            out.println();
            out.println( "void " + classname + "::" + "changeState( " + stateClassname + " *newState ) {" );
            out.println( "    m_state = newState;" );
            out.println( "}" );
            out.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }


    private void generateState() throws IOException {
        try {
            String classname = m_basename + "State";
            String fsmClassname = m_basename + "FSM";
            String filename = classname + ".h";

            PrintWriter out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, false );
            out.println();
            out.println();
            out.println( "#ifndef " + classname + "_H" );
            out.println( "#define " + classname + "_H" );
            out.println();
            out.println();
            out.println( "#include \"LogicError.h\"" );
            out.println();
            out.println();
            out.println( "class " + fsmClassname + ";" );
            out.println();
            out.println();
            out.println( "class " + classname );
            out.println( "{" );
            out.println();
            out.println();
            out.println( "    protected:" );
            out.println( "        " + classname + "() {}" );
            out.println();
            out.println();
            out.println( "    public:" );
            out.println( "        virtual ~" + classname + "() {}" );
            out.println();
            out.println();

            Iterator iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println( "    virtual void " + methodname + "( " + fsmClassname + " *ctx, " + m_eparamType + " o ) throw (LogicError) = 0;" );
            }

            out.println( "};" );

            String stateClassname = classname;

            Iterator it = m_states.iterator();
            while ( it.hasNext() ) {
                String cstate = (String)it.next();
                classname = m_basename + cstate + "State";

//                if ( m_verbose )
//                    System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
                out.println();
                out.println();
                out.println( "class " + classname + " : public " + stateClassname );
                out.println( "{" );
                out.println( "    private:" );
                out.println( "        static " + classname + " *m_instance;" );
                out.println(); out.println();
                out.println( "    private:" );
                out.println( "        " + classname + "();" );
                out.println( "        virtual ~" + classname + "();" );
                out.println(); out.println();
                out.println( "    public:" );
                out.println( "        static " + classname + "* instance();" );
                out.println(); out.println();

                iter = m_entries.iterator();
                while ( iter.hasNext() ) {
                    SMTEntry entry = (SMTEntry)iter.next();
                    if ( cstate.equals( entry.m_actualState ) ) {
//                        String nextCstate = m_basename + entry.m_nextState + "State";
                        out.println( "    virtual void " + entry.m_event + "( " + fsmClassname + " *ctx,  " + m_eparamType + " o ) throw (LogicError);" );
                    }
                }
                out.println( "};" );
            }

            out.println();
            out.println();
            out.println( "#endif" );
            out.close();


            // now generate the implementation
            classname = m_basename + "State";
            fsmClassname = m_basename + "FSM";
            filename = classname + ".cxx";

            out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, false );
            out.println();
            out.println();
            out.println( "#include \"" + classname + ".h\"" );
            out.println( "#include \"" + fsmClassname + ".h\"" );
            out.println();
            out.println();

            stateClassname = classname;

            it = m_states.iterator();
            while ( it.hasNext() ) {
                String cstate = (String)it.next();
                classname = m_basename + cstate + "State";

//                if ( m_verbose )
//                    System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
                out.println();
                out.println();
                out.println( classname + " * " + classname + "::m_instance = 0;" );
                out.println();
                out.println();
                out.println( classname + "::" + classname + "() {}" );
                out.println();
                out.println();
                out.println( classname + "::~" + classname + "() {}" );
                out.println();
                out.println();
                out.println( classname + " * " + classname + "::instance() {" );
                out.println( "    if ( m_instance == 0 )" );
                out.println( "        m_instance = new " + classname + ";" );
                out.println( "    return m_instance;" );
                out.println( "}" );

                iter = m_entries.iterator();
                while ( iter.hasNext() ) {
                    SMTEntry entry = (SMTEntry)iter.next();
                    if ( cstate.equals( entry.m_actualState ) ) {
                        String nextCstate = m_basename + entry.m_nextState + "State";
                        String errorCstate = m_basename + m_errorState + "State";
                        out.println();
                        out.println();
                        out.println( "void " + classname + "::" + entry.m_event + "( " + fsmClassname + " *ctx,  " + m_eparamType + " o ) throw (LogicError) {" );
                        if ( "-".equals( entry.m_action ) ) {
                            if ( !"-".equals( entry.m_nextState ) )
                                out.println( "    ctx->changeState( " + nextCstate + "::instance() );" );
                            out.println( "}" );
                        }
                        else if ( "!".equals( entry.m_action ) ) {
                            out.println( "    ctx->changeState( " + errorCstate + "::instance() );" );
                            out.println( "    throw LogicError();" );
                            out.println( "}" );
                        }
                        else {
                            out.println( "    try {" );
                            out.println( "        ctx->" + entry.m_action + "( o );" );
                            out.println( "    }" );
                            out.println( "    catch( LogicError &e ) {" );
                            out.println( "        ctx->changeState( " + errorCstate + "::instance() );" );
                            out.println( "        throw;" );
                            out.println( "    }" );
                            if ( !"-".equals( entry.m_nextState ) )
                                out.println( "        ctx->changeState( " + nextCstate + "::instance() );" );
                            out.println( "}" );
                        }
                    }
                }
            }

            out.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }


    private void writeHeader( PrintWriter out, boolean editable ) {
        if ( m_copy != null )
            out.println( m_copy );
        out.println( "/*" );
        if ( !editable )
            out.println( " * this file was generated by NunniFSMGen - do not edit!" );
        else
            out.println( " * this file was generated by NunniFSMGen" );
        out.println( " */" );
    }
}
