/*
 * $Id: FSMGen.java,v 1.10 2004/02/23 15:15:18 nunnari Exp $
 *
 * Copyright (c) 2003-2004 Roberto Nunnari - roberto.nunnari@nunnisoft.ch
 * All rights reserved.
 *
 *  This file is part of NunniFSMGen.
 *
 *  NunniFSMGen is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  NunniFSMGen is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with NunniFSMGen; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.nunnisoft.fsmgen;

import java.io.*;
import java.util.*;
import com.nunnisoft.fsmgen.lang.*;

/**
 *
 * @author  Roberto Nunnari
 */
public class FSMGen {

    BufferedReader  m_smt;
    String          m_basename = null;
    String          m_initialState = null;
    String          m_errorState = null;
    String          m_package = null;
    String          m_eparamType = null;
    String          m_copy = null;
    Vector          m_entries = null;
    boolean         m_verbose = false;


    /** Creates a new instance of FSMGen */
    public FSMGen( BufferedReader br, boolean verbose ) {
        m_smt = br;
        m_verbose = verbose;
        m_entries = new Vector( 10, 10 );
        m_basename = null;
        m_errorState = null;
        m_package = null;
        m_eparamType = null;
        m_copy = null;

        String line = "";
        try {
            while ( ( line = m_smt.readLine() ) != null ) {

                line = line.trim();

                if ( line.equals( "" ) )
                    ;
                else if ( line.startsWith( "Context" ) ) {
                    String[] w = line.split( "\\s+" );
                    m_basename = w[1];
                }
                else if ( line.startsWith( "InitialState" ) ) {
                    String[] w = line.split( "\\s+" );
                    m_initialState = w[1];
                }
                else if ( line.startsWith( "ErrorState" ) ) {
                    String[] w = line.split( "\\s+" );
                    m_errorState = w[1];
                }
                else if ( line.startsWith( "Package" ) ) {
                    String[] w = line.split( "\\s+" );
                    m_package = w[1];
                }
                else if ( line.startsWith( "EventParamType" ) ) {
                    String[] w = line.split( "\\s+" );
                    m_eparamType = line.substring( w[0].length() ).trim();
                }
                else if ( line.startsWith( "Copyright" ) ) {
                    String[] w = line.split( "\\s+" );
                    m_copy = line.substring( w[0].length() ).trim();
                }
                else {
                    String s[] = line.split( "\\s+" );
                    for ( int i = 0; i < s.length; ++i )
                        s[i] = s[i].trim();

                    SMTEntry entry = new SMTEntry( s[0], s[1], s[2], s[3] );
                    m_entries.add( entry );
                }
            }
            if ( m_copy != null ) {
                StringBuffer sb = new StringBuffer();
                BufferedReader copyreader = new BufferedReader( new FileReader( m_copy ) );
                while ( ( line = copyreader.readLine() ) != null ) {
                    sb.append( line ).append( "\r\n" );
                }
                m_copy = sb.toString();
            }

            if ( verbose ) {
                System.out.println( "From configuration file:" );
                System.out.println( "  Context is: " + m_basename );
                System.out.println( "  InitialState is: " + m_initialState );
                System.out.println( "  ErrorState is: " + m_errorState );
                System.out.println( "  Package is: " + m_package );
                System.out.println( "  EventParamType is: " + m_eparamType );
                System.out.println( "  Copyright file is: " + m_copy );
                System.out.println( "  here are the FSM transitions found:" );
                Iterator iter = m_entries.iterator();
                while ( iter.hasNext() ) {
                    SMTEntry entry = (SMTEntry)iter.next();
                    System.out.println( "    " + entry.m_actualState + ", " + entry.m_event + ", " + entry.m_nextState + ", " + entry.m_action );
                }
            }
        }
        catch ( IOException e ) {
            e.printStackTrace();
        }
    }

    public Generator createGenerator( String language, String genPack, String genName ) {
        Generator generator = null;

        // controllo parametri dal file di configurazione
        if ( m_basename == null )
            m_basename = "Context";
        if ( m_errorState == null )
            m_errorState = "ERROR";
        if ( m_package == null )
            m_package = "generated";

        // controllo parametri command line
        if ( genPack == null || "".equals( genPack ) )
            genPack = m_package;
        if ( genName == null || "".equals( genName ) )
            genName = m_basename;

        if ( "java".equalsIgnoreCase( language ) ) {
            if ( m_eparamType == null )
                m_eparamType = "Object";
            generator = new JavaGenerator( genName, genPack, m_eparamType, m_copy, m_initialState, m_errorState, m_entries, m_verbose );
        }
        else if ( "c".equalsIgnoreCase( language ) ) {
            if ( m_eparamType == null )
                m_eparamType = "void *";
            generator = new CGenerator( genName, m_eparamType, m_copy, m_initialState, m_errorState, m_entries, m_verbose );
        }
        else if ( "c++".equalsIgnoreCase( language ) || "cxx".equalsIgnoreCase( language ) || "cpp".equalsIgnoreCase( language ) ) {
            if ( m_eparamType == null )
                m_eparamType = "void *";
            generator = new CppGenerator( genName, m_eparamType, m_copy, m_initialState, m_errorState, m_entries, m_verbose );
        }
        else {
            System.out.println( language + " " + ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.IS_NOT_VALID_LANGUAGE ) );
            usage();
        }
        return generator;
    }


    private static void showProgramInfo() {
        System.out.println( "Copyright (c) 2003 Roberto Nunnari - roberto.nunnari@nunnisoft.ch" );
        System.out.println( "All rights reserved." );
        System.out.println();
        System.out.println( "This program is free software; you can redistribute it and/or modify" );
        System.out.println( "it under the terms of the GNU General Public License as published by" );
        System.out.println( "the Free Software Foundation; either version 2 of the License, or" );
        System.out.println( "(at your option) any later version." );
        System.out.println();
        System.out.println( "This program is distributed in the hope that it will be useful," );
        System.out.println( "but WITHOUT ANY WARRANTY; without even the implied warranty of" );
        System.out.println( "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the" );
        System.out.println( "GNU General Public License for more details." );
        System.out.println();
        System.out.println( "You should have received a copy of the GNU General Public License" );
        System.out.println( "along with NunniFSMGen; if not, write to the Free Software" );
        System.out.println( "Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA" );
    }

    private static void showVersion() {
        System.out.println(
            ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.VERSION )
            + " " + Version.getVersion() );
    }


    private static void usage() {
        System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.USAGE ) );
    }

    public static void main( String[] args ) {
        if ( args.length < 1 ) {
            usage();
            System.exit( 1 );
        }
       
        String filename = args[0];
        boolean verbose = false;

        final int none = 0;
        final int lang = 1;
        final int pack = 2;
        final int name = 3;
        int state = none;
        String genLang = "java";
        String genPack = null;
        String genName = null;
        for ( int i = 0; i < args.length; ++i ) {
            filename = args[i];
            if ( state == lang ) {
                state = none;
                genLang = args[i];
            }
            else if ( state == pack ) {
                state = none;
                genPack = args[i];
            }
            else if ( state == name ) {
                state = none;
                genName = args[i];
            }
            else if ( "-v".equals( args[i] ) )
                verbose = true;
            else if ( "-o".equals( args[i] ) )
                state = lang;
            else if ( "-p".equals( args[i] ) )
                state = pack;
            else if ( "-n".equals( args[i] ) )
                state = name;
            else if ( "-?".equals( args[i] ) ) {
                usage();
                System.exit( 0 );
            }
            else if ( "-V".equals( args[i] ) ) {
                showVersion();
                System.exit( 0 );
            }
            else if ( "-i".equals( args[i] ) ) {
                showProgramInfo();
                System.exit( 0 );
            }
            else if ( args[i].startsWith( "-" ) ) {
                System.out.println( args[i] + " " + ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.IS_NOT_VALID_OPTION ) );
                usage();
                System.exit( -1 );
            }
        }

        try {
            BufferedReader br = new BufferedReader( new FileReader( filename ) );
            FSMGen fsmGen = new FSMGen( br, verbose );
            Generator generator = fsmGen.createGenerator( genLang, genPack, genName );
            if ( generator == null )
                System.exit( -1 );
            generator.generate();
        }
        catch ( FileNotFoundException e ) {
            e.printStackTrace();
            System.exit( 1 );
        }
        catch ( IOException e ) {
            e.printStackTrace();
            System.exit( 1 );
        }
        System.exit( 0 );
    }
}
