/*
 * $Id: JavaGenerator.java,v 1.8 2004/02/08 22:53:20 nunnari Exp $
 *
 * Copyright (c) 2003-2004 Roberto Nunnari - roberto.nunnari@nunnisoft.ch
 * All rights reserved.
 *
 *  This file is part of NunniFSMGen.
 *
 *  NunniFSMGen is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  NunniFSMGen is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with NunniFSMGen; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.nunnisoft.fsmgen;

import java.io.*;
import java.util.*;
import com.nunnisoft.fsmgen.lang.*;

/**
 *
 * @author  Roberto Nunnari
 */
public class JavaGenerator implements Generator {

    String          m_basename = null;
    String          m_copy = null;
    String          m_genPack = null;
    String          m_eparamType = null;
    String          m_initialState = null;
    String          m_errorState = null;
    Vector          m_entries = null;
    Vector          m_states = null;
    Vector          m_events = null;
    Vector          m_contextMethods = null;
    boolean         m_verbose = false;


    /** Creates a new instance of JavaGenerator */
    public JavaGenerator( String baseName, String genPack, String eparamType, String copy, String initialState, String errorState, Vector entries, boolean verbose ) {
        m_basename = baseName;
        m_genPack = genPack;
        m_eparamType = eparamType;
        if ( copy == null )
            m_copy = copy;
        else
            m_copy = copy.trim();
        m_initialState = initialState;
        m_errorState = errorState;
        m_verbose = verbose;
        initFromTransitions( entries );
    }

    private void initFromTransitions( Vector entries ) {
        m_entries = entries;
        m_states = new Vector( 10, 10 );
        m_events = new Vector( 10, 10 );
        m_contextMethods = new Vector( 10, 10 );
        Iterator iter = entries.iterator();
        while ( iter.hasNext() ) {
            SMTEntry entry = (SMTEntry)iter.next();

            if ( !m_states.contains( entry.m_actualState ) )
                m_states.add( entry.m_actualState );
            if ( !m_events.contains( entry.m_event ) )
                m_events.add( entry.m_event );
            if ( !m_contextMethods.contains( entry.m_action ) )
                m_contextMethods.add( entry.m_action );
        }
    }

    public void generate() throws IOException {
        if ( m_verbose )
            System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GEN_JAVA_SRC ) );
        generateExceptionClass();
        generateContext();
        generateFSM();
        generateState();
        generateConcreteStates();
    }



    private void generateExceptionClass() throws IOException {
        try {
            String filename = "LogicError.java";
            PrintWriter out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, false );
            out.println();
            out.println();
            out.println( "package " + m_genPack + ";" );
            out.println();
            out.println();
            out.println( "public class LogicError extends Exception {" );
            out.println( "    public LogicError() {" );
            out.println( "        super();" );
            out.println( "    }" );
            out.println();
            out.println();
            out.println( "    public LogicError( String message ) {" );
            out.println( "        super( message );" );
            out.println( "    }" );
            out.println( "}" );
            out.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }


    private void generateContext() throws IOException {
        try {
            String classname = m_basename;
            String filename = classname + ".java";

            File file = new File( filename );
            if ( file.exists() ) {
                System.out.print( "File '" + filename + "' " + ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.ALREADY_EXISTS ) + ": " );
                classname += "New";
                filename = classname + ".java";
                System.out.println( filename );
            }
            PrintWriter out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, true );
            out.println();
            out.println();
            out.println( "package " + m_genPack + ";" );
            out.println();
            out.println();
            out.println( "abstract class " + classname );
            out.println( "{" );

            Iterator iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println( "    public abstract void " + methodname + "( " + m_eparamType + " o ) throws LogicError;" );
            }

            out.println();
            out.println();
            out.println( "    protected " + classname + "() {}" );
            out.println();
            out.println();
            out.println( "    protected " + classname + "( Object o ) {}" );

            iter = m_contextMethods.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                if ( !"-".equals( methodname ) && !"!".equals( methodname ) ) {
                    out.println();
                    out.println();
                    out.println( "    boolean " + methodname + "( " + m_eparamType + " o ) throws LogicError {" );
                    out.println( "        // add your code hereafter..." );
                    out.println( "        return true;" );
                    out.println( "    }" );
                }
            }

            out.println( "}" );
            out.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }


    private void generateFSM() throws IOException {
        try {
            String classname = m_basename + "FSM";
            String stateClassname = m_basename + "State";
            String filename = classname + ".java";

            PrintWriter out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, false );
            out.println();
            out.println();
            out.println( "package " + m_genPack + ";" );
            out.println();
            out.println();
            out.println( "public class " + classname + " extends " + m_basename );
            out.println( "{" );
            out.println( "    private " + stateClassname + " m_state = null;" );
            out.println();
            out.println();
            out.println( "    private " + classname + "() {}" );
            out.println();
            out.println();
            out.println( "    public " + classname + "( Object o ) {" );
            out.println( "        super( o );" );
            out.println( "        m_state = "  + m_basename + m_initialState + "State.instance();" );
            out.println( "    }" );
            out.println();
            out.println();
            out.println( "    public String toString() {" );
            out.println( "        return m_state.toString();" );
            out.println( "    }" );


            Iterator iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println();
                out.println();
                out.println( "    public void " + methodname + "( " + m_eparamType + " o ) throws LogicError {" );
                out.println( "        m_state." + methodname + "( this, o );" );
                out.println( "    }" );
            }

            out.println();
            out.println();
            out.println( "    void changeState( " + stateClassname + " newState ) {" );
            out.println( "        m_state = newState;" );
            out.println( "    }" );
            out.println( "}" );
            out.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }


    private void generateState() throws IOException {
        try {
            String classname = m_basename + "State";
            String fsmClassname = m_basename + "FSM";
            String filename = classname + ".java";

            PrintWriter out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
            if ( m_verbose )
                System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
            writeHeader( out, false );
            out.println();
            out.println();
            out.println( "package " + m_genPack + ";" );
            out.println();
            out.println();
            out.println( "class " + classname );
            out.println( "{" );
            out.println( "    " + classname + "() {}" );
            out.println();
            out.println();

            Iterator iter = m_events.iterator();
            while ( iter.hasNext() ) {
                String methodname = (String)iter.next();
                out.println( "    void " + methodname + "( " + fsmClassname + " ctx, " + m_eparamType + " o ) throws LogicError {}" );
            }

            out.println( "}" );
            out.close();
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }


    private void generateConcreteStates() throws IOException {
        String stateClassname = m_basename + "State";
        String fsmClassname = m_basename + "FSM";


        try {
            PrintWriter out = null;
            Iterator it = m_states.iterator();
            while ( it.hasNext() ) {
                String cstate = (String)it.next();
                String classname = m_basename + cstate + "State";
                String filename = classname + ".java";

                out = new PrintWriter( new BufferedWriter( new FileWriter( filename ) ) );
                if ( m_verbose )
                    System.out.println( ResourceBundle.getBundle( "com.nunnisoft.fsmgen.lang.GlobalRes" ).getString( GlobalRes.GENERATING ) + " " + filename );
                writeHeader( out, false );
                out.println();
                out.println();
                out.println( "package " + m_genPack + ";" );
                out.println();
                out.println();
                out.println( "class " + classname + " extends " + stateClassname );
                out.println( "{" );
                out.println( "    private static " + classname + " m_instance = null;" );
                out.println(); out.println();
                out.println( "    private " + classname + "() {}" );
                out.println(); out.println();
                out.println( "    static " + classname + " instance() {" );
                out.println( "        if ( m_instance == null )" );
                out.println( "            m_instance = new " + classname + "();" );
                out.println( "        return m_instance;" );
                out.println( "    }" );

                Iterator iter = m_entries.iterator();
                while ( iter.hasNext() ) {
                    SMTEntry entry = (SMTEntry)iter.next();
                    if ( cstate.equals( entry.m_actualState ) ) {
                        String nextCstate = m_basename + entry.m_nextState + "State";
                        String errorCstate = m_basename + m_errorState + "State";
                        out.println(); out.println();
                        out.println( "    void " + entry.m_event + "( " + fsmClassname + " ctx,  " + m_eparamType + " o ) throws LogicError {" );
                        if ( "-".equals( entry.m_action ) ) {
                            if ( !"-".equals( entry.m_nextState ) )
                                out.println( "        ctx.changeState( " + nextCstate + ".instance() );" );
                            out.println( "    }" );
                        }
                        else if ( "!".equals( entry.m_action ) ) {
                            out.println( "        ctx.changeState( " + errorCstate + ".instance() );" );
                            out.println( "        throw new LogicError();" );
                            out.println( "    }" );
                        }
                        else {
                            out.println( "        try {" );
                            out.println( "            ctx." + entry.m_action + "( o );" );
                            out.println( "        }" );
                            out.println( "        catch( LogicError e ) {" );
                            out.println( "            ctx.changeState( " + errorCstate + ".instance() );" );
                            out.println( "            throw e;" );
                            out.println( "        }" );
                            if ( !"-".equals( entry.m_nextState ) )
                                out.println( "        ctx.changeState( " + nextCstate + ".instance() );" );
                            out.println( "    }" );
                        }
                    }
                }
                out.println( "}" );
                out.close();
            }
        }
        catch ( IOException e ) {
            e.printStackTrace();
            throw e;
        }
    }

    private void writeHeader( PrintWriter out, boolean editable ) {
        if ( m_copy != null )
            out.println( m_copy );
        out.println( "/*" );
        if ( !editable )
            out.println( " * this file was generated by NunniFSMGen - do not edit!" );
        else
            out.println( " * this file was generated by NunniFSMGen" );
        out.println( " */" );
    }
}
